/*
Read a .PLY file
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <arpa/inet.h>
#include <ctype.h>
#include <unistd.h>


#include "vector.h"
#include "bvol.h"
#include "world.h"
#include "object.h"
#include "surface.h"
#include "sphere.h"
#include "triangle.h"
#include "disc.h"
#include "read_ply.h"
#include "vertex.h"





inline int mymatch(char *a, char *b)
{
 return strncmp(a, b, strlen(b));
}

inline int mymatch2(char *a, char *b, char **c)
{
 int m = strncmp(a, b, strlen(b));
 if(c != NULL) 
    {
     if(!m)  
         *c = a + strlen(b);
        else
         *c = a;
    }
 return !m;
}

inline int mymatch3(char *a, char *b, char **c)
{
 int i=0, m;
 while(a[i] && isspace(a[i])) i++;
 m = strncmp(a + i, b, strlen(b));
 if(c != NULL) 
    {
     if(!m)  
         *c = a + strlen(b) + i;
        else
         *c = a + i;
    }
 return !m;
}



typedef enum {PLY_FORMAT_UNKNOWN=0, PLY_FORMAT_ASCII, PLY_FORMAT_LITTLE, PLY_FORMAT_BIG, PLY_FORMAT_COUNT} ply_format;
char *ply_format_name[] = {"unknown", "ascii", "binary_little_endian", "binary_big_endian"};
char *ply_format_name2[] = {"UNKNOWN!", "ASCII", "little-endian binary", "big-endian binary"};
ply_format ply_format_fromname(char *n, char **mt)
{
 ply_format t;
 for(t = 0; t < PLY_FORMAT_COUNT; t++)
     if(mymatch3(n, ply_format_name[t], mt))
         return t;
 fprintf(stderr, "PLY format %s not recognized!\n", n);
 return PLY_FORMAT_UNKNOWN;
}



typedef enum {PLY_TYPE_UNKNOWN=0, PLY_TYPE_CHAR, PLY_TYPE_UCHAR, PLY_TYPE_SHORT, PLY_TYPE_USHORT, PLY_TYPE_INT, PLY_TYPE_UINT, PLY_TYPE_FLOAT, PLY_TYPE_DOUBLE, 
              PLY_TYPE_LIST, PLY_TYPE_COUNT} ply_type;
int ply_type_size[] = {0, 1, 1, 2, 2, 4, 4, 4, 8};
char *ply_type_name[] = {"unknown", "char", "uchar", "short", "ushort", "int", "uint", "float", "double", "list"};

ply_type ply_type_fromname(char *n, char **mt)
{
 ply_type t;
 for(t = 0; t < PLY_TYPE_COUNT; t++)
     if(mymatch3(n, ply_type_name[t], mt))
         return t;
 fprintf(stderr, "PLY type %s not recognized!\n", n);
 return PLY_TYPE_UNKNOWN;
}



typedef enum {PLY_ELEM_NONE=0, PLY_ELEM_VERT, PLY_ELEM_FACE, PLY_ELEM_TRISTRIP, PLY_ELEM_COUNT} ply_elem;
char *ply_elem_name[] = {"none", "vertex", "face", "tristrips"};

ply_elem ply_elem_fromname(char *n, char **mt)
{
 ply_elem t;
 for(t = 0; t < PLY_ELEM_COUNT; t++)
     if(mymatch3(n, ply_elem_name[t], mt))
         return t;
 fprintf(stderr, "PLY element %s not recognized!\n", n);
 return PLY_ELEM_NONE;
}


typedef enum {PLY_PROP_UNKNOWN=0, 
              PLY_PROP_X, PLY_PROP_Y, PLY_PROP_Z,             /* vertex position */
              PLY_PROP_VERTEX_INDICES,                        /* list of verticies for a face */
              PLY_PROP_VNX, PLY_PROP_VNY, PLY_PROP_VNZ,       /* vertex normal */
              PLY_PROP_TRISTRIP_INDICES,                      /* list of verticies for a triangle strip */
              PLY_PROP_COUNT} ply_prop;

char *ply_prop_name[] = {"\x00unknown",
              "\x01x", "\x01y", "\x01z",
              "\x02vertex_indices",
              "\x01nx", "\x01ny", "\x01nz",
              "\x03vertex_indices"};

ply_type ply_prop_fromname(char *n, ply_elem e, char **mt)
{
 ply_prop t;
 for(t = 0; t < PLY_PROP_COUNT; t++)
     if((e == ply_prop_name[t][0]) && mymatch3(n, ply_prop_name[t] + 1, mt))
         return t;
 fprintf(stderr, "PLY property %s not recognized for element %s!\n", n, ply_elem_name[e]);
 return PLY_PROP_UNKNOWN;
}



int endmatch(char *s, char *m)
{
 if(strlen(m) > strlen(s)) return 0;
 return !strcmp(s + (strlen(s) - strlen(m)), m);
}



double ply_read(FILE *f, ply_type type, ply_format format);
void ply_normalize(vertex *vertxs, int numverts, vectype scalegoal);
void ply_gennormals(vertex *vertxs, int numverts, int **faces, int numfaces);
void ply_eofcheck(FILE *f);


/* no element may have more than this many properties.  should be more than sufficient. */
#define PLY_MAXPROPS 100
void read_ply(world w, char *filename)
{
 int i, facenum;
 int **faces = NULL, **strips = NULL;
 vertex *vertxs = NULL;
 char buf[1024], *mt;
 obj_triangle tri;
 surface surf;
 obj ob;
 FILE *f;

/* int axis[] = {0, 2, -1}; */
 int axis[] = {0, 1, 2}; 
 vectype scalegoal = 20;

 ply_format format = PLY_FORMAT_UNKNOWN;
 ply_elem elements[PLY_ELEM_COUNT], element = PLY_ELEM_NONE;
 ply_type types[PLY_ELEM_COUNT][PLY_MAXPROPS];  /* zero isn't used, but oh well. */
 ply_prop props[PLY_ELEM_COUNT][PLY_MAXPROPS];
 int numprops[PLY_ELEM_COUNT], numitems[PLY_ELEM_COUNT], numelems = 0, elementnum;

 int readvertnorms = 0;

 for(i = 0; i < PLY_ELEM_COUNT; i++)
     numprops[i] = numitems[i] = 0;


 if(endmatch(filename, ".bz2"))
    {
     char buf[1024];
     snprintf(buf, 1000, "bzcat \"%s\" >/tmp/foo.tmp.ply", filename);
     fprintf(stderr, "PLY decompression: %s\n", buf);
     system(buf);
     f = fopen("/tmp/foo.tmp.ply", "r");
    }
    else
    {
     f = fopen(filename, "r");
    }
 if(!f)
    {
     fprintf(stderr, "Unable to open %s!\n", filename);
     return;
    }
 fgets(buf, 1000, f);
 if(!mymatch2(buf, "ply", NULL))
    {
     fprintf(stderr, "PLY header not found, not a PLY file?\n");
     return;
    }

 while(!feof(f) && fgets(buf, 1000, f))
    {
     if(!strtok(buf, "\r\n")) continue;
     if(mymatch2(buf, "end_header", NULL))
        {
         fprintf(stderr, "PLY header end.\n");
         break;
        }
     if(mymatch2(buf, "comment ", &mt))
        {
         fprintf(stderr, "PLY file comment: %s\n", mt);
         continue;
        }

     if(mymatch2(buf, "format ", &mt))
        {
         format = ply_format_fromname(mt, &mt);
         fprintf(stderr, "PLY file format is %s, version %s.\n", ply_format_name2[format], mt + 1);
         if(!mymatch3(mt, "1.0", NULL))
            {
             fprintf(stderr, "PLY version is not 1.0, file may not be read correctly.\n");
            }
         continue;
        }

     if(mymatch2(buf, "element ", &mt))
        {
         element = ply_elem_fromname(mt, &mt);
         for(i = 0; i < numelems; i++)
             if(elements[i] == element)
                {
                 fprintf(stderr, "Element %s already found!\n", ply_elem_name[element]);
                 return;
                }
         numitems[element] = atoi(mt);
         elements[numelems] = element;
         numelems++;
         fprintf(stderr, "PLY element %s has %d item(s).\n", ply_elem_name[element], numitems[element]);
         continue;
        }

     if(mymatch2(buf, "property ", &mt))
        {
         ply_prop p;
         ply_type t;
         if(element <= PLY_ELEM_NONE || element >= PLY_ELEM_COUNT)
            {
             fprintf(stderr, "PLY property found with invalid/unknown element!\n");
             continue;
            }
         if(numprops[element] >= PLY_MAXPROPS - 4)
            {
             fprintf(stderr, "Too many properties for element %s!\n", ply_elem_name[element]);
             return;
            }
         t = ply_type_fromname(mt, &mt);
         if(t != PLY_TYPE_LIST)
            {
             p = ply_prop_fromname(mt, element, &mt);
             types[element][numprops[element]] = t;
             props[element][numprops[element]] = p;
             numprops[element]++;
             fprintf(stderr, "PLY %s property %s, type %s.\n", ply_elem_name[element], ply_prop_name[p]+1, ply_type_name[t]);
            }
            else
            {
             ply_type t2;
             types[element][numprops[element]] = PLY_TYPE_LIST;
             types[element][numprops[element] + 1] = t = ply_type_fromname(mt, &mt);
             types[element][numprops[element] + 2] = t2 = ply_type_fromname(mt, &mt);
             p = ply_prop_fromname(mt, element, &mt);
             props[element][numprops[element]] = p;
             props[element][numprops[element] + 1] = p;
             props[element][numprops[element] + 2] = p;
             numprops[element] += 3;
             fprintf(stderr, "PLY %s list %s, index type %s, data type %s.\n", ply_elem_name[element], ply_prop_name[p]+1, ply_type_name[t], ply_type_name[t2]);
            }
         continue;
        }
     fprintf(stderr, "Unknown line in PLY header '%s'!\n", buf);
    }

 if(feof(f))
    {
     fprintf(stderr, "PLY file EOF before end of header!  Not a PLY file?\n");
     return;
    }
 if(format == PLY_FORMAT_UNKNOWN)
    {
     fprintf(stderr, "End of PLY header without recognizing file format!\n");
     return;
    }

/*
 if(numitems[PLY_ELEM_VERT] > 0) verts = malloc(sizeof(vector) * numitems[PLY_ELEM_VERT]);
 if(numitems[PLY_ELEM_VERT] > 0) vertnorms = malloc(sizeof(vector) * numitems[PLY_ELEM_VERT]);
*/

 if(numitems[PLY_ELEM_VERT] > 0) 
    {
     vertxs = malloc(sizeof(vertex) * numitems[PLY_ELEM_VERT]);
     if(vertxs)
         for(i = 0; i < numitems[PLY_ELEM_VERT]; i++)
             vertxs[i] = vertex_new();
    }

 if(numitems[PLY_ELEM_FACE] > 0) faces = (int **)malloc(sizeof(int *) * numitems[PLY_ELEM_FACE]);
 if(numitems[PLY_ELEM_TRISTRIP] > 0) strips = (int **)malloc(sizeof(int *) * numitems[PLY_ELEM_TRISTRIP]);
/* if( (numitems[PLY_ELEM_VERT]&&!vertxs) || (numitems[PLY_ELEM_VERT]&&!vertnorms) || (numitems[PLY_ELEM_FACE]&&!faces) || (numitems[PLY_ELEM_TRISTRIP]&&!strips) ) */
 if( (numitems[PLY_ELEM_VERT]&&!vertxs) || (numitems[PLY_ELEM_FACE]&&!faces) || (numitems[PLY_ELEM_TRISTRIP]&&!strips) )
    {
     fprintf(stderr, "Error allocating storage for PLY elements!\n");
     return;
    }


 for(elementnum = 0; elementnum < numelems; elementnum++)
    {
     int itemnum, propnum;
     double data;
     ply_prop p;
     element = elements[elementnum];
     for(itemnum = 0; itemnum < numitems[element]; itemnum++)
        {
         if(feof(f))
            {
             fprintf(stderr, "End of file while reading element %s!\n", ply_elem_name[element]);
             return;
            }
         propnum = 0;
         while(propnum < numprops[element])
            {
             if(types[element][propnum] != PLY_TYPE_LIST)
                {
                 p = props[element][propnum];
                 data = ply_read(f, types[element][propnum], format);
                 /* we could load the entire file into a data structure then
                 fetch things from it later, but for now, this is simpler. */
                 switch(p) 
                    {
                     case PLY_PROP_X:
                     case PLY_PROP_Y:
                     case PLY_PROP_Z:
                         vertxs[itemnum]->point[abs(axis[p - PLY_PROP_X])] = data * ( (axis[p - PLY_PROP_X] < 0) ? -1.0 : 1.0 );
                         break;
                     case PLY_PROP_VNX:
                     case PLY_PROP_VNY:
                     case PLY_PROP_VNZ:
                         vertxs[itemnum]->normal[abs(axis[p - PLY_PROP_VNX])] = data * ( (axis[p - PLY_PROP_VNX] < 0) ? -1.0 : 1.0 );
                         vertex_mark_normal(vertxs[itemnum]);
                         readvertnorms++;
                         break;
                     default:
                         break;
                         /* fprintf(stderr, "."); */
                    }
                 propnum++;
                }
                else
                {
                 int listitem, listcount = (int)(ply_read(f, types[element][propnum+1], format));
                 if(listcount < 0)
                    {
                     fprintf(stderr, "List count may not be negative!\n");
                     return;
                    }
                 p = props[element][propnum];
                 switch(p)
                    {
                     case PLY_PROP_VERTEX_INDICES:
                         faces[itemnum] = (int *)malloc(sizeof(int) * (listcount + 1));
                         faces[itemnum][0] = listcount;
                         break;
                     case PLY_PROP_TRISTRIP_INDICES:
                         strips[itemnum] = (int *)malloc(sizeof(int) * (listcount + 1));
                         strips[itemnum][0] = listcount;
                         break;
                     default:
                         break;
                         /* fprintf(stderr, ","); */
                    }
                 for(listitem = 0; listitem < listcount; listitem++)
                    {
                     int v;
                     data = ply_read(f, types[element][propnum+2], format);
                     switch(p)
                        {
                         case PLY_PROP_VERTEX_INDICES:
                             v = (int)data;
                             if(v < 0 || v >= numitems[PLY_ELEM_VERT])
                                {
                                 fprintf(stderr, "PLY file referrs to non-existant vertex #%d!\n", v);
                                 return;
                                }
                             faces[itemnum][listitem + 1] = v;
                             break;
                         case PLY_PROP_TRISTRIP_INDICES:
                             v = (int)data;
                             if((v < 0 || v >= numitems[PLY_ELEM_VERT]) && v != -1)
                                {
                                 fprintf(stderr, "PLY file referrs to non-existant vertex #%d!\n", v);
                                 return;
                                }
                             strips[itemnum][listitem + 1] = v;
                             break;
                         default:
                             break;
                             /* fprintf(stderr, "'"); */
                        }
                    }
                 propnum += 3;
                }
            }
        }
     fprintf(stderr, "PLY element %s read with %d items.\n", ply_elem_name[element], itemnum);
    }



 ply_eofcheck(f);


 if(numitems[PLY_ELEM_TRISTRIP])
    {
     int newpoints=0, newfaces=0, stripnum, vertnum, dir=0;
     for(stripnum = 0; stripnum < numitems[PLY_ELEM_TRISTRIP]; stripnum++)
        {
         faces = (int **)realloc(faces, sizeof(int *) * (numitems[PLY_ELEM_FACE] + strips[stripnum][0] + 1));
         newpoints += strips[stripnum][0];
         for(vertnum = 3; vertnum < strips[stripnum][0]; vertnum++)
            {
             if(strips[stripnum][vertnum] < 0)
                {
                 vertnum += 2;
                 dir = 0;
                 continue;
                }
             faces[numitems[PLY_ELEM_FACE]] = (int *)malloc(sizeof(int) * (3 + 1));
             faces[numitems[PLY_ELEM_FACE]][0] = 3;
             for(i = 0; i < 3; i++)
                {
                 if(strips[stripnum][vertnum - 2 + i] < 0)
                    {
                     fprintf(stderr, "Invalid tristrip!  not enough points to form a triangle...\n");
                     return;
                    }
                 if(!dir)
                     faces[numitems[PLY_ELEM_FACE]][i + 1] = strips[stripnum][vertnum - 2 + i];
                    else
                     faces[numitems[PLY_ELEM_FACE]][i + 1] = strips[stripnum][vertnum - i];
                }
             dir = !dir;
             newfaces++;
             numitems[PLY_ELEM_FACE]++;
            }
        }

     fprintf(stderr, "PLY tristrips with %d listed verticies converted to %d faces.\n", newpoints, newfaces);
    }


 if(strips)
    {
     for(i = 0; i < numitems[PLY_ELEM_TRISTRIP]; i++)
         free(strips[i]);
     free(strips);
    }

 fprintf(stderr, "PLY file %d faces using %d verticies read.\n", numitems[PLY_ELEM_FACE], numitems[PLY_ELEM_VERT]);
 if(scalegoal != 0.0) ply_normalize(vertxs, numitems[PLY_ELEM_VERT], scalegoal);
 if(!readvertnorms) ply_gennormals(vertxs, numitems[PLY_ELEM_VERT], faces, numitems[PLY_ELEM_FACE]);


 surf = surface_new_set_rgbdsp(drand() / 2.0 + 0.5, drand() / 2.0 + 0.5, drand() / 2.0 + 0.5, 1.0, 1.0, 10.0);
 if(1) surf = surface_new_set_mirror();
 for(facenum = 0; facenum < numitems[PLY_ELEM_FACE]; facenum++)
    {
     if(faces[facenum][0] != 3)
        {
         fprintf(stderr, "PLY currently may only contain triangles, but %d verticies found!\n", faces[facenum][0]);
         continue;
        }
     tri = obj_triangle_new_set_basic(vertxs[faces[facenum][1]], vertxs[faces[facenum][2]], vertxs[faces[facenum][3]]);
     if(1) {vector 
n;triangle_compute_normal(n,vertxs[faces[facenum][1]]->point,vertxs[faces[facenum][2]]->point,vertxs[faces[facenum][3]]->point);surf=surface_new_set_rgbdsp(fabs(n[1]),fabs(n[0]),fabs(n[2]),1.0,1.0,10.0);}
     if(0) surf = surface_new_set_rgbdsp(drand() / 2.0 + 0.5, drand() / 2.0 + 0.5, drand() / 2.0 + 0.5, 1.0, 1.0, 10.0);
     if(0) if(rand()%2) surf = surface_new_set_mirror();
     ob = obj_new_set(obj_type_triangle, tri);
     obj_set_surface(ob, surf);
     world_add_obj(w, ob);
    }


 if(vertxs) free(vertxs);  /* don't free actual verticies! */
 if(faces)
    {
     for(facenum = 0; facenum < numitems[PLY_ELEM_FACE]; facenum++)
         free(faces[facenum]);
     free(faces);
    }
 fclose(f);

 fprintf(stderr, "PLY file done.\n");
}





void ply_gennormals(vertex *vertxs, int numverts, int **faces, int numfaces)
{
 int vertnum, facenum, i;
 vector normal = {0, 0, 0};
 for(vertnum = 0; vertnum < numverts; vertnum++)
     for(i = 0; i < 3; i++)
         vertex_set_normal(vertxs[vertnum], normal);

 for(facenum = 0; facenum < numfaces; facenum++)
    {
     if(faces[facenum][0] < 3)
         continue;

     /* we assume the object is planar.  if not...  tough luck! */
     triangle_compute_normal(normal, vertxs[faces[facenum][1]]->point, vertxs[faces[facenum][2]]->point, vertxs[faces[facenum][3]]->point);
     if(!(normal[0] >= -100 && normal[0] <= 100))
         continue;

     for(i = 1; i <= faces[facenum][0]; i++)
        {
         vector_add_3d(vertxs[faces[facenum][i]]->normal, vertxs[faces[facenum][i]]->normal, normal);
        }
    }

 for(vertnum = 0; vertnum < numverts; vertnum++)
     vector_norm_3d(vertxs[vertnum]->normal, vertxs[vertnum]->normal);


 fprintf(stderr, "PLY normals computed.\n");
}





void ply_normalize(vertex *vertxs, int numverts, vectype scalegoal)
{
 vectype scale = 1, d;
 vector min={0,0,0}, max={0,0,0}, offset={0,0,0};
 int vertnum, i;

 for(vertnum = 0; vertnum < numverts; vertnum++)
     for(i = 0; i < 3; i++)
        {
         d = vertxs[vertnum]->point[i];
         if(d < min[i] || vertnum == 0) min[i] = d;
         if(d > max[i] || vertnum == 0) max[i] = d;
        }
 fprintf(stderr, "PLY extremes (%f - %f), (%f - %f), (%f - %f).\n", min[0], max[0], min[1], max[1], min[2], max[2]);

 scale = vtmin3(scalegoal / (max[0] - min[0]), scalegoal / (max[1] - min[1]), scalegoal / (max[2] - min[2]));
 for(i = 0; i < 3; i++)
     offset[i] = 0.0 - (max[i] + min[i]) * scale / 2.0;
 for(vertnum = 0; vertnum < numverts; vertnum++)
     for(i = 0; i < 3; i++)
         vertxs[vertnum]->point[i] = vertxs[vertnum]->point[i] * scale + offset[i];

 fprintf(stderr, "PLY scaled by %f, offset by (%f, %f, %f).\n", scale, offset[0], offset[1], offset[2]);
}




/* double can accurately store all official .ply data types.  Reading of
ASCII format is very slow, and should be improved with a buffer. */
double ply_read(FILE *f, ply_type type, ply_format format)
{
 unsigned char buf[128], buf2[8];
 int rb, i;
 ply_format hostformat = (ntohs(12345) == 12345) ? PLY_FORMAT_BIG : PLY_FORMAT_LITTLE;
 if(type == PLY_TYPE_UNKNOWN || type >= PLY_TYPE_COUNT)
    {
     fprintf(stderr, "Attempt to read an unknown .PLY data type!\n");
     return 0;
    }
 if(format == PLY_FORMAT_LITTLE || format == PLY_FORMAT_BIG)
    {
     if(fread(buf, 1, ply_type_size[type], f) != ply_type_size[type])
        {
         fprintf(stderr, "Error reading %d bytes from .PLY file!\n", ply_type_size[type]);
         return 0;
        }
     if(format != hostformat)
        {
         for(i = 0; i < ply_type_size[type]; i++)
             buf2[i] = buf[ply_type_size[type] - 1 - i];
        }
        else
        {
         for(i = 0; i < ply_type_size[type]; i++)
             buf2[i] = buf[i];
        }
     switch(type)
       {
        case PLY_TYPE_CHAR:
            return *((signed char *)(buf2));
        case PLY_TYPE_UCHAR:
            return *((unsigned char *)(buf2));
        case PLY_TYPE_SHORT:
            return *((signed short *)(buf2));
        case PLY_TYPE_USHORT:
            return *((unsigned short *)(buf2));
        case PLY_TYPE_INT:
            return *((signed int *)(buf2));
        case PLY_TYPE_UINT:
            return *((unsigned int *)(buf2));
        case PLY_TYPE_FLOAT:
            return *((float *)(buf2));
        case PLY_TYPE_DOUBLE:
            return *((double *)(buf2));
        default:
            fprintf(stderr, "Unable to convert unknown .PLY data type!\n");
            return 0;
       }
    }
    else  /* ASCII */
    {
     rb = 0;
     while(1)
        {
         if(!fread(buf+rb, 1, 1, f))
            {
             if(rb == 0)
                {
                 fprintf(stderr, "Error reading value from .PLY file!\n");
                 return 0;
                }
             break;
            }
         if(!isdigit(buf[rb]) && buf[rb] != '-' && buf[rb] != '.' && buf[rb] != 'e' && buf[rb] != '+')
            {
             if(!isspace(buf[rb]))
                {
                 fprintf(stderr, "Unexpected character '%c' (0x%x) reading .PLY file!\n", buf[rb], buf[rb]);
                 return 0;
                }
             if(rb == 0) continue;  else  break;
            }
         rb++;
         if(rb >= 128)
            {
             fprintf(stderr, "Unable to locate value in .PLY file, too many bytes read!\n");
             return 0;
            }
        }
     buf[rb] = '\0';
     /* fprintf(stderr, "%f ", atof((char *)buf)); */
     return atof((char *)buf);
    }

 return 0.0;
}



void ply_eofcheck(FILE *f)
{
 char buf[1024];
 while(fgets(buf, 1000, f))
    {
     if(!strtok(buf, "\r\n\t ")) continue;
     fprintf(stderr, "PLY contains information past expected end of file!\n");
     /* fprintf(stderr, "PLY input: %s\n", buf); */
     break;
    }
}
